@@ -79,11 +79,26 @@ bool recorded(const std::chrono::steady_clock::time_point &t) {
 } // namespace
 
 Config::Config()
-    : data_length(-1), addrs(nullptr), nreqs(1), nclients(1), nthreads(1),
-      max_concurrent_streams(-1), window_bits(30), connection_window_bits(30),
-      rate(0), rate_period(1.0), conn_active_timeout(0.),
-      conn_inactivity_timeout(0.), no_tls_proto(PROTO_HTTP2), data_fd(-1),
-      port(0), default_port(0), verbose(false), timing_script(false) {}
+    : data_length(-1),
+      addrs(nullptr),
+      nreqs(1),
+      nclients(1),
+      nthreads(1),
+      max_concurrent_streams(1),
+      window_bits(30),
+      connection_window_bits(30),
+      rate(0),
+      rate_period(1.0),
+      conn_active_timeout(0.),
+      conn_inactivity_timeout(0.),
+      no_tls_proto(PROTO_HTTP2),
+      data_fd(-1),
+      port(0),
+      default_port(0),
+      verbose(false),
+      timing_script(false),
+      base_uri_unix(false),
+      unix_addr{} {}
 
 Config::~Config() {
   if (base_uri_unix) {
@@ -106,9 +121,18 @@ constexpr size_t MAX_SAMPLES = 1000000;
 } // namespace
 
 Stats::Stats(size_t req_todo, size_t nclients)
-    : req_todo(req_todo), req_started(0), req_done(0), req_success(0),
-      req_status_success(0), req_failed(0), req_error(0), req_timedout(0),
-      bytes_total(0), bytes_head(0), bytes_head_decomp(0), bytes_body(0),
+    : req_todo(req_todo),
+      req_started(0),
+      req_done(0),
+      req_success(0),
+      req_status_success(0),
+      req_failed(0),
+      req_error(0),
+      req_timedout(0),
+      bytes_total(0),
+      bytes_head(0),
+      bytes_head_decomp(0),
+      bytes_body(0),
       status() {}
 
 Stream::Stream() : req_stat{}, status_success(-1) {}
@@ -178,7 +202,9 @@ void readcb(struct ev_loop *loop, ev_io *w, int revents) {
   auto client = static_cast<Client *>(w->data);
   client->restart_timeout();
   if (client->do_read() != 0) {
-    client->fail();
+    if (client->try_again_or_fail() == 0) {
+      return;
+    }
     delete client;
     return;
   }
@@ -290,9 +316,18 @@ void client_request_timeout_cb(struct ev_loop *loop, ev_timer *w, int revents) {
 } // namespace
 
 Client::Client(uint32_t id, Worker *worker, size_t req_todo)
-    : cstat{}, worker(worker), ssl(nullptr), next_addr(config.addrs),
-      current_addr(nullptr), reqidx(0), state(CLIENT_IDLE), req_todo(req_todo),
-      req_started(0), req_done(0), id(id), fd(-1),
+    : cstat{},
+      worker(worker),
+      ssl(nullptr),
+      next_addr(config.addrs),
+      current_addr(nullptr),
+      reqidx(0),
+      state(CLIENT_IDLE),
+      req_todo(req_todo),
+      req_started(0),
+      req_done(0),
+      id(id),
+      fd(-1),
       new_connection_requested(false) {
   ev_io_init(&wev, writecb, 0, EV_WRITE);
   ev_io_init(&rev, readcb, 0, EV_READ);
@@ -420,7 +455,7 @@ void Client::restart_timeout() {
   }
 }
 
-void Client::fail() {
+int Client::try_again_or_fail() {
   disconnect();
 
   if (new_connection_requested) {
@@ -438,13 +473,21 @@ void Client::fail() {
 
       // Keep using current address
       if (connect() == 0) {
-        return;
+        return 0;
       }
       std::cerr << "client could not connect to host" << std::endl;
     }
   }
 
   process_abandoned_streams();
+
+  return -1;
+}
+
+void Client::fail() {
+  disconnect();
+
+  process_abandoned_streams();
 }
 
 void Client::disconnect() {
@@ -727,9 +770,10 @@ int Client::connection_made() {
 #endif // OPENSSL_VERSION_NUMBER >= 0x10002000L
 
     if (next_proto) {
-      if (util::check_h2_is_selected(next_proto, next_proto_len)) {
+      auto proto = StringRef{next_proto, next_proto_len};
+      if (util::check_h2_is_selected(proto)) {
         session = make_unique<Http2Session>(this);
-      } else if (util::streq_l(NGHTTP2_H1_1, next_proto, next_proto_len)) {
+      } else if (util::streq(NGHTTP2_H1_1, proto)) {
         session = make_unique<Http1Session>(this);
       }
 #ifdef HAVE_SPDYLAY
@@ -743,20 +787,18 @@ int Client::connection_made() {
 
       // Just assign next_proto to selected_proto anyway to show the
       // negotiation result.
-      selected_proto.assign(next_proto, next_proto + next_proto_len);
+      selected_proto = proto.str();
     } else {
       std::cout << "No protocol negotiated. Fallback behaviour may be activated"
                 << std::endl;
 
       for (const auto &proto : config.npn_list) {
-        if (std::equal(NGHTTP2_H1_1_ALPN,
-                       NGHTTP2_H1_1_ALPN + str_size(NGHTTP2_H1_1_ALPN),
-                       proto.c_str())) {
+        if (util::streq(NGHTTP2_H1_1_ALPN, StringRef{proto})) {
           std::cout
               << "Server does not support NPN/ALPN. Falling back to HTTP/1.1."
               << std::endl;
           session = make_unique<Http1Session>(this);
-          selected_proto = NGHTTP2_H1_1;
+          selected_proto = NGHTTP2_H1_1.str();
           break;
         }
       }
@@ -784,7 +826,7 @@ int Client::connection_made() {
       break;
     case Config::PROTO_HTTP1_1:
       session = make_unique<Http1Session>(this);
-      selected_proto = NGHTTP2_H1_1;
+      selected_proto = NGHTTP2_H1_1.str();
       break;
 #ifdef HAVE_SPDYLAY
     case Config::PROTO_SPDY2:
@@ -1094,11 +1136,20 @@ void Client::try_new_connection() { new_connection_requested = true; }
 
 Worker::Worker(uint32_t id, SSL_CTX *ssl_ctx, size_t req_todo, size_t nclients,
                size_t rate, size_t max_samples, Config *config)
-    : stats(req_todo, nclients), loop(ev_loop_new(0)), ssl_ctx(ssl_ctx),
-      config(config), id(id), tls_info_report_done(false),
-      app_info_report_done(false), nconns_made(0), nclients(nclients),
-      nreqs_per_client(req_todo / nclients), nreqs_rem(req_todo % nclients),
-      rate(rate), max_samples(max_samples), next_client_id(0) {
+    : stats(req_todo, nclients),
+      loop(ev_loop_new(0)),
+      ssl_ctx(ssl_ctx),
+      config(config),
+      id(id),
+      tls_info_report_done(false),
+      app_info_report_done(false),
+      nconns_made(0),
+      nclients(nclients),
+      nreqs_per_client(req_todo / nclients),
+      nreqs_rem(req_todo % nclients),
+      rate(rate),
+      max_samples(max_samples),
+      next_client_id(0) {
   if (!config->is_rate_mode()) {
     progress_interval = std::max(static_cast<size_t>(1), req_todo / 10);
   } else {
@@ -1344,7 +1395,7 @@ std::string get_reqline(const char *uri, const http_parser_url &u) {
   std::string reqline;
 
   if (util::has_uri_field(u, UF_PATH)) {
-    reqline = util::get_uri_field(uri, u, UF_PATH);
+    reqline = util::get_uri_field(uri, u, UF_PATH).str();
   } else {
     reqline = "/";
   }
@@ -1378,15 +1429,15 @@ constexpr char UNIX_PATH_PREFIX[] = "unix:";
 } // namespace
 
 namespace {
-bool parse_base_uri(std::string base_uri) {
+bool parse_base_uri(const StringRef &base_uri) {
   http_parser_url u{};
   if (http_parser_parse_url(base_uri.c_str(), base_uri.size(), 0, &u) != 0 ||
       !util::has_uri_field(u, UF_SCHEMA) || !util::has_uri_field(u, UF_HOST)) {
     return false;
   }
 
-  config.scheme = util::get_uri_field(base_uri.c_str(), u, UF_SCHEMA);
-  config.host = util::get_uri_field(base_uri.c_str(), u, UF_HOST);
+  config.scheme = util::get_uri_field(base_uri.c_str(), u, UF_SCHEMA).str();
+  config.host = util::get_uri_field(base_uri.c_str(), u, UF_HOST).str();
   config.default_port = util::get_default_port(base_uri.c_str(), u);
   if (util::has_uri_field(u, UF_PORT)) {
     config.port = u.port;
@@ -1411,7 +1462,7 @@ std::vector<std::string> parse_uris(std::vector<std::string>::iterator first,
 
   if (!config.has_base_uri()) {
 
-    if (!parse_base_uri(*first)) {
+    if (!parse_base_uri(StringRef{*first})) {
       std::cerr << "invalid URI: " << *first << std::endl;
       exit(EXIT_FAILURE);
     }
@@ -1802,24 +1853,27 @@ int main(int argc, char **argv) {
     case 'i':
       config.ifile = optarg;
       break;
-    case 'p':
-      if (util::strieq(NGHTTP2_CLEARTEXT_PROTO_VERSION_ID, optarg)) {
+    case 'p': {
+      auto proto = StringRef{optarg};
+      if (util::strieq(StringRef::from_lit(NGHTTP2_CLEARTEXT_PROTO_VERSION_ID),
+                       proto)) {
         config.no_tls_proto = Config::PROTO_HTTP2;
-      } else if (util::strieq(NGHTTP2_H1_1, optarg)) {
+      } else if (util::strieq(NGHTTP2_H1_1, proto)) {
         config.no_tls_proto = Config::PROTO_HTTP1_1;
 #ifdef HAVE_SPDYLAY
-      } else if (util::strieq("spdy/2", optarg)) {
+      } else if (util::strieq_l("spdy/2", proto)) {
         config.no_tls_proto = Config::PROTO_SPDY2;
-      } else if (util::strieq("spdy/3", optarg)) {
+      } else if (util::strieq_l("spdy/3", proto)) {
         config.no_tls_proto = Config::PROTO_SPDY3;
-      } else if (util::strieq("spdy/3.1", optarg)) {
+      } else if (util::strieq_l("spdy/3.1", proto)) {
         config.no_tls_proto = Config::PROTO_SPDY3_1;
 #endif // HAVE_SPDYLAY
       } else {
-        std::cerr << "-p: unsupported protocol " << optarg << std::endl;
+        std::cerr << "-p: unsupported protocol " << proto << std::endl;
         exit(EXIT_FAILURE);
       }
       break;
+    }
     case 'r':
       config.rate = strtoul(optarg, nullptr, 10);
       if (config.rate == 0) {
@@ -1845,18 +1899,19 @@ int main(int argc, char **argv) {
       }
       break;
     case 'B': {
+      auto arg = StringRef{optarg};
       config.base_uri = "";
       config.base_uri_unix = false;
 
-      if (util::istarts_with_l(optarg, UNIX_PATH_PREFIX)) {
+      if (util::istarts_with_l(arg, UNIX_PATH_PREFIX)) {
         // UNIX domain socket path
         sockaddr_un un;
 
-        auto path = optarg + str_size(UNIX_PATH_PREFIX);
-        auto pathlen = strlen(optarg) - str_size(UNIX_PATH_PREFIX);
+        auto path = StringRef{std::begin(arg) + str_size(UNIX_PATH_PREFIX),
+                              std::end(arg)};
 
-        if (pathlen == 0 || pathlen + 1 > sizeof(un.sun_path)) {
-          std::cerr << "--base-uri: invalid UNIX domain socket path: " << optarg
+        if (path.size() == 0 || path.size() + 1 > sizeof(un.sun_path)) {
+          std::cerr << "--base-uri: invalid UNIX domain socket path: " << arg
                     << std::endl;
           exit(EXIT_FAILURE);
         }
@@ -1864,18 +1919,19 @@ int main(int argc, char **argv) {
         config.base_uri_unix = true;
 
         auto &unix_addr = config.unix_addr;
-        std::copy_n(path, pathlen + 1, unix_addr.sun_path);
+        std::copy(std::begin(path), std::end(path), unix_addr.sun_path);
+        unix_addr.sun_path[path.size()] = '\0';
         unix_addr.sun_family = AF_UNIX;
 
         break;
       }
 
-      if (!parse_base_uri(optarg)) {
-        std::cerr << "--base-uri: invalid base URI: " << optarg << std::endl;
+      if (!parse_base_uri(arg)) {
+        std::cerr << "--base-uri: invalid base URI: " << arg << std::endl;
         exit(EXIT_FAILURE);
       }
 
-      config.base_uri = optarg;
+      config.base_uri = arg.str();
       break;
     }
     case 'v':
@@ -1904,7 +1960,7 @@ int main(int argc, char **argv) {
         break;
       case 4:
         // npn-list option
-        config.npn_list = util::parse_config_str_list(optarg);
+        config.npn_list = util::parse_config_str_list(StringRef{optarg});
         break;
       case 5:
         // rate-period
@@ -1916,7 +1972,8 @@ int main(int argc, char **argv) {
         break;
       case 6:
         // --h1
-        config.npn_list = util::parse_config_str_list("http/1.1");
+        config.npn_list =
+            util::parse_config_str_list(StringRef::from_lit("http/1.1"));
         config.no_tls_proto = Config::PROTO_HTTP1_1;
         break;
       }
@@ -1940,7 +1997,8 @@ int main(int argc, char **argv) {
   }
 
   if (config.npn_list.empty()) {
-    config.npn_list = util::parse_config_str_list(DEFAULT_NPN_LIST);
+    config.npn_list =
+        util::parse_config_str_list(StringRef::from_lit(DEFAULT_NPN_LIST));
   }
 
   // serialize the APLN tokens
