@@ -114,8 +114,6 @@ void writecb(struct ev_loop *loop, ev_io *w, int revents) {
 int ClientHandler::noop() { return 0; }
 
 int ClientHandler::read_clear() {
-  ev_timer_again(conn_.loop, &conn_.rt);
-
   for (;;) {
     if (rb_.rleft() && on_read() != 0) {
       return -1;
@@ -124,12 +122,24 @@ int ClientHandler::read_clear() {
       rb_.reset();
     } else if (rb_.wleft() == 0) {
       conn_.rlimit.stopw();
+      if (reset_conn_rtimer_required_) {
+        reset_conn_rtimer_required_ = false;
+        ev_timer_again(conn_.loop, &conn_.rt);
+      }
+      return 0;
+    }
+
+    if (!ev_is_active(&conn_.rev)) {
       return 0;
     }
 
     auto nread = conn_.read_clear(rb_.last, rb_.wleft());
 
     if (nread == 0) {
+      if (reset_conn_rtimer_required_) {
+        reset_conn_rtimer_required_ = false;
+        ev_timer_again(conn_.loop, &conn_.rt);
+      }
       return 0;
     }
 
@@ -204,8 +214,6 @@ int ClientHandler::tls_handshake() {
 }
 
 int ClientHandler::read_tls() {
-  ev_timer_again(conn_.loop, &conn_.rt);
-
   ERR_clear_error();
 
   for (;;) {
@@ -217,12 +225,26 @@ int ClientHandler::read_tls() {
       rb_.reset();
     } else if (rb_.wleft() == 0) {
       conn_.rlimit.stopw();
+      if (reset_conn_rtimer_required_) {
+        reset_conn_rtimer_required_ = false;
+        ev_timer_again(conn_.loop, &conn_.rt);
+      }
+
+      return 0;
+    }
+
+    if (!ev_is_active(&conn_.rev)) {
       return 0;
     }
 
     auto nread = conn_.read_tls(rb_.last, rb_.wleft());
 
     if (nread == 0) {
+      if (reset_conn_rtimer_required_) {
+        reset_conn_rtimer_required_ = false;
+        ev_timer_again(conn_.loop, &conn_.rt);
+      }
+
       return 0;
     }
 
@@ -369,22 +391,22 @@ int ClientHandler::upstream_http1_connhd_read() {
 }
 
 ClientHandler::ClientHandler(Worker *worker, int fd, SSL *ssl,
-                             const char *ipaddr, const char *port)
+                             const char *ipaddr, const char *port, int family,
+                             const UpstreamAddr *faddr)
     : conn_(worker->get_loop(), fd, ssl, worker->get_mcpool(),
             get_config()->conn.upstream.timeout.write,
             get_config()->conn.upstream.timeout.read,
             get_config()->conn.upstream.ratelimit.write,
             get_config()->conn.upstream.ratelimit.read, writecb, readcb,
             timeoutcb, this, get_config()->tls.dyn_rec.warmup_threshold,
-            get_config()->tls.dyn_rec.idle_timeout),
-      pinned_http2sessions_(
-          get_config()->conn.downstream.proto == PROTO_HTTP2
-              ? make_unique<std::vector<ssize_t>>(
-                    get_config()->conn.downstream.addr_groups.size(), -1)
-              : nullptr),
-      ipaddr_(ipaddr), port_(port), worker_(worker),
+            get_config()->tls.dyn_rec.idle_timeout, PROTO_NONE),
+      ipaddr_(ipaddr),
+      port_(port),
+      faddr_(faddr),
+      worker_(worker),
       left_connhd_len_(NGHTTP2_CLIENT_MAGIC_LEN),
-      should_close_after_write_(false) {
+      should_close_after_write_(false),
+      reset_conn_rtimer_required_(false) {
 
   ++worker_->get_worker_stat()->num_connections;
 
@@ -406,11 +428,19 @@ ClientHandler::ClientHandler(Worker *worker, int fd, SSL *ssl,
 
   auto &fwdconf = get_config()->http.forwarded;
 
-  if ((fwdconf.params & FORWARDED_FOR) &&
-      fwdconf.for_node_type == FORWARDED_NODE_OBFUSCATED) {
-    forwarded_for_obfuscated_ = "_";
-    forwarded_for_obfuscated_ += util::random_alpha_digit(
-        worker_->get_randgen(), SHRPX_OBFUSCATED_NODE_LENGTH);
+  if (fwdconf.params & FORWARDED_FOR) {
+    if (fwdconf.for_node_type == FORWARDED_NODE_OBFUSCATED) {
+      forwarded_for_ = "_";
+      forwarded_for_ += util::random_alpha_digit(worker_->get_randgen(),
+                                                 SHRPX_OBFUSCATED_NODE_LENGTH);
+    } else if (family == AF_INET6) {
+      forwarded_for_ = "[";
+      forwarded_for_ += ipaddr_;
+      forwarded_for_ += ']';
+    } else {
+      // family == AF_INET or family == AF_UNIX
+      forwarded_for_ = ipaddr_;
+    }
   }
 }
 
@@ -482,6 +512,10 @@ void ClientHandler::reset_upstream_write_timeout(ev_tstamp t) {
   }
 }
 
+void ClientHandler::signal_reset_upstream_conn_rtimer() {
+  reset_conn_rtimer_required_ = true;
+}
+
 int ClientHandler::validate_next_proto() {
   const unsigned char *next_proto = nullptr;
   unsigned int next_proto_len;
@@ -514,26 +548,26 @@ int ClientHandler::validate_next_proto() {
     return 0;
   }
 
+  auto proto = StringRef{next_proto, next_proto_len};
+
   if (LOG_ENABLED(INFO)) {
-    std::string proto(next_proto, next_proto + next_proto_len);
     CLOG(INFO, this) << "The negotiated next protocol: " << proto;
   }
 
-  if (!ssl::in_proto_list(get_config()->tls.npn_list, next_proto,
-                          next_proto_len)) {
+  if (!ssl::in_proto_list(get_config()->tls.npn_list, proto)) {
     if (LOG_ENABLED(INFO)) {
-      CLOG(INFO, this) << "The negotiated protocol is not supported";
+      CLOG(INFO, this) << "The negotiated protocol is not supported: " << proto;
     }
     return -1;
   }
 
-  if (util::check_h2_is_selected(next_proto, next_proto_len)) {
+  if (util::check_h2_is_selected(proto)) {
     on_read_ = &ClientHandler::upstream_http2_connhd_read;
 
     auto http2_upstream = make_unique<Http2Upstream>(this);
 
     upstream_ = std::move(http2_upstream);
-    alpn_.assign(next_proto, next_proto + next_proto_len);
+    alpn_.assign(std::begin(proto), std::end(proto));
 
     // At this point, input buffer is already filled with some bytes.
     // The read callback is not called until new data come. So consume
@@ -546,7 +580,7 @@ int ClientHandler::validate_next_proto() {
   }
 
 #ifdef HAVE_SPDYLAY
-  auto spdy_version = spdylay_npn_get_version(next_proto, next_proto_len);
+  auto spdy_version = spdylay_npn_get_version(proto.byte(), proto.size());
   if (spdy_version) {
     upstream_ = make_unique<SpdyUpstream>(spdy_version, this);
 
@@ -575,9 +609,9 @@ int ClientHandler::validate_next_proto() {
   }
 #endif // HAVE_SPDYLAY
 
-  if (next_proto_len == 8 && memcmp("http/1.1", next_proto, 8) == 0) {
+  if (proto == StringRef::from_lit("http/1.1")) {
     upstream_ = make_unique<HttpsUpstream>(this);
-    alpn_ = "http/1.1";
+    alpn_ = proto.str();
 
     // At this point, input buffer is already filled with some bytes.
     // The read callback is not called until new data come. So consume
@@ -622,13 +656,18 @@ void ClientHandler::pool_downstream_connection(
   if (!dconn->poolable()) {
     return;
   }
+
+  dconn->set_client_handler(nullptr);
+
+  auto group = dconn->get_downstream_addr_group();
+
   if (LOG_ENABLED(INFO)) {
     CLOG(INFO, this) << "Pooling downstream connection DCONN:" << dconn.get()
-                     << " in group " << dconn->get_group();
+                     << " in group " << group;
   }
-  dconn->set_client_handler(nullptr);
-  auto dconn_pool = worker_->get_dconn_pool();
-  dconn_pool->add_downstream_connection(std::move(dconn));
+
+  auto &dconn_pool = group->shared_addr->dconn_pool;
+  dconn_pool.add_downstream_connection(std::move(dconn));
 }
 
 void ClientHandler::remove_downstream_connection(DownstreamConnection *dconn) {
@@ -636,51 +675,58 @@ void ClientHandler::remove_downstream_connection(DownstreamConnection *dconn) {
     CLOG(INFO, this) << "Removing downstream connection DCONN:" << dconn
                      << " from pool";
   }
-  auto dconn_pool = worker_->get_dconn_pool();
-  dconn_pool->remove_downstream_connection(dconn);
+  auto &dconn_pool =
+      dconn->get_downstream_addr_group()->shared_addr->dconn_pool;
+  dconn_pool.remove_downstream_connection(dconn);
 }
 
 std::unique_ptr<DownstreamConnection>
 ClientHandler::get_downstream_connection(Downstream *downstream) {
-  size_t group;
+  size_t group_idx;
   auto &downstreamconf = get_config()->conn.downstream;
-  auto &groups = downstreamconf.addr_groups;
   auto catch_all = downstreamconf.addr_group_catch_all;
+  auto &groups = worker_->get_downstream_addr_groups();
 
   const auto &req = downstream->request();
 
   // Fast path.  If we have one group, it must be catch-all group.
-  // HTTP/2 and client proxy modes fall in this case.
+  // proxy mode falls in this case.
   if (groups.size() == 1) {
-    group = 0;
+    group_idx = 0;
   } else if (req.method == HTTP_CONNECT) {
     //  We don't know how to treat CONNECT request in host-path
     //  mapping.  It most likely appears in proxy scenario.  Since we
     //  have dealt with proxy case already, just use catch-all group.
-    group = catch_all;
+    group_idx = catch_all;
   } else {
     auto &router = get_config()->router;
+    auto &wildcard_patterns = get_config()->wildcard_patterns;
     if (!req.authority.empty()) {
-      group = match_downstream_addr_group(router, req.authority, req.path,
-                                          groups, catch_all);
+      group_idx =
+          match_downstream_addr_group(router, wildcard_patterns, req.authority,
+                                      req.path, groups, catch_all);
     } else {
       auto h = req.fs.header(http2::HD_HOST);
       if (h) {
-        group = match_downstream_addr_group(router, h->value, req.path, groups,
-                                            catch_all);
+        group_idx = match_downstream_addr_group(
+            router, wildcard_patterns, h->value, req.path, groups, catch_all);
       } else {
-        group = match_downstream_addr_group(router, "", req.path, groups,
-                                            catch_all);
+        group_idx =
+            match_downstream_addr_group(router, wildcard_patterns, StringRef{},
+                                        req.path, groups, catch_all);
       }
     }
   }
 
   if (LOG_ENABLED(INFO)) {
-    CLOG(INFO, this) << "Downstream address group: " << group;
+    CLOG(INFO, this) << "Downstream address group_idx: " << group_idx;
   }
 
-  auto dconn_pool = worker_->get_dconn_pool();
-  auto dconn = dconn_pool->pop_downstream_connection(group);
+  auto &group = worker_->get_downstream_addr_groups()[group_idx];
+  auto &shared_addr = group.shared_addr;
+  auto &dconn_pool = shared_addr->dconn_pool;
+
+  auto dconn = dconn_pool.pop_downstream_connection();
 
   if (!dconn) {
     if (LOG_ENABLED(INFO)) {
@@ -688,22 +734,45 @@ ClientHandler::get_downstream_connection(Downstream *downstream) {
                        << " Create new one";
     }
 
-    auto dconn_pool = worker_->get_dconn_pool();
+    if (shared_addr->proto == PROTO_HTTP2) {
+      auto &http2_freelist = shared_addr->http2_freelist;
 
-    if (downstreamconf.proto == PROTO_HTTP2) {
       Http2Session *http2session;
-      auto &pinned = (*pinned_http2sessions_)[group];
-      if (pinned == -1) {
-        http2session = worker_->next_http2_session(group);
-        pinned = http2session->get_index();
+
+      if (http2_freelist.empty() ||
+          http2_freelist.size() < shared_addr->addrs.size()) {
+        if (LOG_ENABLED(INFO)) {
+          if (http2_freelist.empty()) {
+            CLOG(INFO, this)
+                << "http2_freelist is empty; create new Http2Session";
+          } else {
+            CLOG(INFO, this) << "Create new Http2Session; current "
+                             << http2_freelist.size() << ", min "
+                             << shared_addr->addrs.size();
+          }
+        }
+        http2session = new Http2Session(
+            conn_.loop, shared_addr->tls ? worker_->get_cl_ssl_ctx() : nullptr,
+            worker_, &group);
       } else {
-        auto dgrp = worker_->get_dgrp(group);
-        http2session = dgrp->http2sessions[pinned].get();
+        http2session = http2_freelist.head;
+        http2_freelist.remove(http2session);
       }
-      dconn = make_unique<Http2DownstreamConnection>(dconn_pool, http2session);
+
+      if (http2session->max_concurrency_reached(1)) {
+        if (LOG_ENABLED(INFO)) {
+          CLOG(INFO, this) << "Maximum streams are reached for Http2Session("
+                           << http2session
+                           << "). Remove Http2Session from http2_freelist";
+        }
+      } else {
+        http2_freelist.append(http2session);
+      }
+
+      dconn = make_unique<Http2DownstreamConnection>(http2session);
     } else {
       dconn =
-          make_unique<HttpDownstreamConnection>(dconn_pool, group, conn_.loop);
+          make_unique<HttpDownstreamConnection>(&group, conn_.loop, worker_);
     }
     dconn->set_client_handler(this);
     return dconn;
@@ -723,10 +792,6 @@ MemchunkPool *ClientHandler::get_mcpool() { return worker_->get_mcpool(); }
 
 SSL *ClientHandler::get_ssl() const { return conn_.tls.ssl; }
 
-ConnectBlocker *ClientHandler::get_connect_blocker() const {
-  return worker_->get_connect_blocker();
-}
-
 void ClientHandler::direct_http2_upgrade() {
   upstream_ = make_unique<Http2Upstream>(this);
   alpn_ = NGHTTP2_CLEARTEXT_PROTO_VERSION_ID;
@@ -747,23 +812,6 @@ int ClientHandler::perform_http2_upgrade(HttpsUpstream *http) {
   auto downstream = http->get_downstream();
   auto input = downstream->get_response_buf();
 
-  static constexpr char res[] =
-      "HTTP/1.1 101 Switching Protocols\r\n"
-      "Connection: Upgrade\r\n"
-      "Upgrade: " NGHTTP2_CLEARTEXT_PROTO_VERSION_ID "\r\n"
-      "\r\n";
-
-  auto required_size = str_size(res) + input->rleft();
-
-  if (output->wleft() < required_size) {
-    if (LOG_ENABLED(INFO)) {
-      CLOG(INFO, this)
-          << "HTTP Upgrade failed because of insufficient buffer space: need "
-          << required_size << ", available " << output->wleft();
-    }
-    return -1;
-  }
-
   if (upstream->upgrade_upstream(http) != 0) {
     return -1;
   }
@@ -775,11 +823,15 @@ int ClientHandler::perform_http2_upgrade(HttpsUpstream *http) {
   on_read_ = &ClientHandler::upstream_http2_connhd_read;
   write_ = &ClientHandler::write_clear;
 
-  auto nread =
-      downstream->get_response_buf()->remove(output->last, output->wleft());
-  output->write(nread);
+  input->remove(*output, input->rleft());
 
-  output->write(res, str_size(res));
+  constexpr auto res =
+      StringRef::from_lit("HTTP/1.1 101 Switching Protocols\r\n"
+                          "Connection: Upgrade\r\n"
+                          "Upgrade: " NGHTTP2_CLEARTEXT_PROTO_VERSION_ID "\r\n"
+                          "\r\n");
+
+  output->append(res);
   upstream_ = std::move(upstream);
 
   signal_write();
@@ -788,11 +840,11 @@ int ClientHandler::perform_http2_upgrade(HttpsUpstream *http) {
 
 bool ClientHandler::get_http2_upgrade_allowed() const { return !conn_.tls.ssl; }
 
-std::string ClientHandler::get_upstream_scheme() const {
+StringRef ClientHandler::get_upstream_scheme() const {
   if (conn_.tls.ssl) {
-    return "https";
+    return StringRef::from_lit("https");
   } else {
-    return "http";
+    return StringRef::from_lit("http");
   }
 }
 
@@ -806,24 +858,35 @@ namespace {
 // is mostly same routine found in
 // HttpDownstreamConnection::push_request_headers(), but vastly
 // simplified since we only care about absolute URI.
-std::string construct_absolute_request_uri(const Request &req) {
+StringRef construct_absolute_request_uri(BlockAllocator &balloc,
+                                         const Request &req) {
   if (req.authority.empty()) {
     return req.path;
   }
 
-  std::string uri;
+  auto len = req.authority.size() + req.path.size();
+  if (req.scheme.empty()) {
+    len += str_size("http://");
+  } else {
+    len += req.scheme.size() + str_size("://");
+  }
+
+  auto iov = make_byte_ref(balloc, len + 1);
+  auto p = iov.base;
+
   if (req.scheme.empty()) {
     // We may have to log the request which lacks scheme (e.g.,
     // http/1.1 with origin form).
-    uri += "http://";
+    p = util::copy_lit(p, "http://");
   } else {
-    uri += req.scheme;
-    uri += "://";
+    p = std::copy(std::begin(req.scheme), std::end(req.scheme), p);
+    p = util::copy_lit(p, "://");
   }
-  uri += req.authority;
-  uri += req.path;
+  p = std::copy(std::begin(req.authority), std::end(req.authority), p);
+  p = std::copy(std::begin(req.path), std::end(req.path), p);
+  *p = '\0';
 
-  return uri;
+  return StringRef{iov.base, p};
 }
 } // namespace
 
@@ -832,15 +895,17 @@ void ClientHandler::write_accesslog(Downstream *downstream) {
   const auto &req = downstream->request();
   const auto &resp = downstream->response();
 
+  auto &balloc = downstream->get_block_allocator();
+
   upstream_accesslog(
       get_config()->logging.access.format,
       LogSpec{
-          downstream, StringRef(ipaddr_), http2::to_method_string(req.method),
+          downstream, StringRef{ipaddr_}, http2::to_method_string(req.method),
 
           req.method == HTTP_CONNECT
               ? StringRef(req.authority)
-              : (get_config()->http2_proxy || get_config()->client_proxy)
-                    ? StringRef(construct_absolute_request_uri(req))
+              : get_config()->http2_proxy
+                    ? StringRef(construct_absolute_request_uri(balloc, req))
                     : req.path.empty()
                           ? req.method == HTTP_OPTIONS
                                 ? StringRef::from_lit("*")
@@ -855,8 +920,8 @@ void ClientHandler::write_accesslog(Downstream *downstream) {
           std::chrono::high_resolution_clock::now(), // request_end_time
 
           req.http_major, req.http_minor, resp.http_status,
-          downstream->response_sent_body_length, StringRef(port_),
-          get_config()->conn.listener.port, get_config()->pid,
+          downstream->response_sent_body_length, StringRef(port_), faddr_->port,
+          get_config()->pid,
       });
 }
 
@@ -879,7 +944,7 @@ void ClientHandler::write_accesslog(int major, int minor, unsigned int status,
                          highres_now,  // request_end_time
                          major, minor, // major, minor
                          status, body_bytes_sent, StringRef(port_),
-                         get_config()->conn.listener.port, get_config()->pid,
+                         faddr_->port, get_config()->pid,
                      });
 }
 
@@ -969,23 +1034,23 @@ int ClientHandler::proxy_protocol_read() {
 
   --end;
 
-  constexpr const char HEADER[] = "PROXY ";
+  constexpr auto HEADER = StringRef::from_lit("PROXY ");
 
-  if (static_cast<size_t>(end - rb_.pos) < str_size(HEADER)) {
+  if (static_cast<size_t>(end - rb_.pos) < HEADER.size()) {
     if (LOG_ENABLED(INFO)) {
       CLOG(INFO, this) << "PROXY-protocol-v1: PROXY version 1 ID not found";
     }
     return -1;
   }
 
-  if (!util::streq_l(HEADER, rb_.pos, str_size(HEADER))) {
+  if (!util::streq(HEADER, StringRef{rb_.pos, HEADER.size()})) {
     if (LOG_ENABLED(INFO)) {
       CLOG(INFO, this) << "PROXY-protocol-v1: Bad PROXY protocol version 1 ID";
     }
     return -1;
   }
 
-  rb_.drain(str_size(HEADER));
+  rb_.drain(HEADER.size());
 
   int family;
 
@@ -1120,63 +1185,18 @@ int ClientHandler::proxy_protocol_read() {
   return on_proxy_protocol_finish();
 }
 
-const std::string &ClientHandler::get_forwarded_by() {
+StringRef ClientHandler::get_forwarded_by() const {
   auto &fwdconf = get_config()->http.forwarded;
 
   if (fwdconf.by_node_type == FORWARDED_NODE_OBFUSCATED) {
-    return fwdconf.by_obfuscated;
-  }
-  if (!local_hostport_.empty()) {
-    return local_hostport_;
+    return StringRef(fwdconf.by_obfuscated);
   }
 
-  auto &listenerconf = get_config()->conn.listener;
-
-  // For UNIX domain socket listener, just return empty string.
-  if (listenerconf.host_unix) {
-    return local_hostport_;
-  }
-
-  int rv;
-  sockaddr_union su;
-  socklen_t addrlen = sizeof(su);
-
-  rv = getsockname(conn_.fd, &su.sa, &addrlen);
-  if (rv != 0) {
-    return local_hostport_;
-  }
-
-  char host[NI_MAXHOST];
-  rv = getnameinfo(&su.sa, addrlen, host, sizeof(host), nullptr, 0,
-                   NI_NUMERICHOST);
-  if (rv != 0) {
-    return local_hostport_;
-  }
-
-  if (su.storage.ss_family == AF_INET6) {
-    local_hostport_ = "[";
-    local_hostport_ += host;
-    local_hostport_ += "]:";
-  } else {
-    local_hostport_ = host;
-    local_hostport_ += ':';
-  }
-
-  local_hostport_ += util::utos(listenerconf.port);
-
-  return local_hostport_;
+  return StringRef{faddr_->hostport};
 }
 
-const std::string &ClientHandler::get_forwarded_for() const {
-  if (get_config()->http.forwarded.for_node_type == FORWARDED_NODE_OBFUSCATED) {
-    return forwarded_for_obfuscated_;
-  }
-
-  if (get_config()->conn.listener.host_unix) {
-    return EMPTY_STRING;
-  }
-
-  return ipaddr_;
+StringRef ClientHandler::get_forwarded_for() const {
+  return StringRef{forwarded_for_};
 }
 
 } // namespace shrpx
