@@ -48,7 +48,8 @@ namespace shrpx {
 
 class Http2DownstreamConnection;
 class Worker;
-class ConnectBlocker;
+struct DownstreamAddrGroup;
+struct DownstreamAddr;
 
 struct StreamData {
   StreamData *dlnext, *dlprev;
@@ -57,13 +58,10 @@ struct StreamData {
 
 class Http2Session {
 public:
-  Http2Session(struct ev_loop *loop, SSL_CTX *ssl_ctx,
-               ConnectBlocker *connect_blocker, Worker *worker, size_t group,
-               size_t idx);
+  Http2Session(struct ev_loop *loop, SSL_CTX *ssl_ctx, Worker *worker,
+               DownstreamAddrGroup *group);
   ~Http2Session();
 
-  int check_cert();
-
   // If hard is true, all pending requests are abandoned and
   // associated ClientHandlers will be deleted.
   int disconnect(bool hard = false);
@@ -92,7 +90,7 @@ public:
   int do_read();
   int do_write();
 
-  int on_read();
+  int on_read(const uint8_t *data, size_t datalen);
   int on_write();
 
   int connected();
@@ -102,13 +100,15 @@ public:
   int read_tls();
   int write_tls();
 
-  int downstream_read_proxy();
+  int downstream_read_proxy(const uint8_t *data, size_t datalen);
   int downstream_connect_proxy();
 
-  int downstream_read();
+  int downstream_read(const uint8_t *data, size_t datalen);
   int downstream_write();
 
   int noop();
+  int read_noop(const uint8_t *data, size_t datalen);
+  int write_noop();
 
   void signal_write();
 
@@ -147,17 +147,31 @@ public:
 
   void submit_pending_requests();
 
-  size_t get_addr_idx() const;
+  DownstreamAddr *get_addr() const;
 
-  size_t get_group() const;
-
-  size_t get_index() const;
+  DownstreamAddrGroup *get_downstream_addr_group() const;
 
   int handle_downstream_push_promise(Downstream *downstream,
                                      int32_t promised_stream_id);
   int handle_downstream_push_promise_complete(Downstream *downstream,
                                               Downstream *promised_downstream);
 
+  // Returns number of downstream connections, including pushed
+  // streams.
+  size_t get_num_dconns() const;
+
+  // Returns true if this object is included in freelist.  See
+  // DownstreamAddrGroup object.
+  bool in_freelist() const;
+
+  // Returns true if the maximum concurrency is reached.  In other
+  // words, the number of currently participated streams in this
+  // session is equal or greater than the max concurrent streams limit
+  // advertised by server.  If |extra| is nonzero, it is added to the
+  // number of current concurrent streams when comparing against
+  // server initiated concurrency limit.
+  bool max_concurrency_reached(size_t extra = 0) const;
+
   enum {
     // Disconnected
     DISCONNECTED,
@@ -185,40 +199,37 @@ public:
   };
 
   using ReadBuf = Buffer<8_k>;
-  using WriteBuf = Buffer<32768>;
+
+  Http2Session *dlnext, *dlprev;
 
 private:
   Connection conn_;
+  DefaultMemchunks wb_;
   ev_timer settings_timer_;
   // This timer has 2 purpose: when it first timeout, set
   // connection_check_state_ = CONNECTION_CHECK_REQUIRED.  After
   // connection check has started, this timer is started again and
   // traps PING ACK timeout.
   ev_timer connchk_timer_;
+  // timer to initiate connection.  usually, this fires immediately.
+  ev_timer initiate_connection_timer_;
   DList<Http2DownstreamConnection> dconns_;
   DList<StreamData> streams_;
   std::function<int(Http2Session &)> read_, write_;
-  std::function<int(Http2Session &)> on_read_, on_write_;
+  std::function<int(Http2Session &, const uint8_t *, size_t)> on_read_;
+  std::function<int(Http2Session &)> on_write_;
   // Used to parse the response from HTTP proxy
   std::unique_ptr<http_parser> proxy_htp_;
   Worker *worker_;
-  ConnectBlocker *connect_blocker_;
   // NULL if no TLS is configured
   SSL_CTX *ssl_ctx_;
+  DownstreamAddrGroup *group_;
+  // Address of remote endpoint
+  DownstreamAddr *addr_;
   nghttp2_session *session_;
-  const uint8_t *data_pending_;
-  size_t data_pendinglen_;
-  // index of get_config()->downstream_addrs this object uses
-  size_t addr_idx_;
-  size_t group_;
-  // index inside group, this is used to pin frontend to certain
-  // HTTP/2 backend for better throughput.
-  size_t index_;
   int state_;
   int connection_check_state_;
   bool flow_control_;
-  WriteBuf wb_;
-  ReadBuf rb_;
 };
 
 nghttp2_session_callbacks *create_http2_downstream_callbacks();
