@@ -81,8 +81,7 @@ public:
   virtual int on_downstream_reset(bool no_retry);
   virtual int send_reply(Downstream *downstream, const uint8_t *body,
                          size_t bodylen);
-  virtual int initiate_push(Downstream *downstream, const char *uri,
-                            size_t len);
+  virtual int initiate_push(Downstream *downstream, const StringRef &uri);
   virtual int response_riovec(struct iovec *iov, int iovcnt) const;
   virtual void response_drain(size_t n);
   virtual bool response_empty() const;
@@ -112,52 +111,26 @@ public:
   void check_shutdown();
 
   int prepare_push_promise(Downstream *downstream);
-  int submit_push_promise(const std::string &scheme,
-                          const std::string &authority, const std::string &path,
-                          Downstream *downstream);
+  int submit_push_promise(const StringRef &scheme, const StringRef &authority,
+                          const StringRef &path, Downstream *downstream);
 
   int on_request_headers(Downstream *downstream, const nghttp2_frame *frame);
 
-  using WriteBuffer = Buffer<32_k>;
-
-  WriteBuffer *get_response_buf();
-
-  void set_pending_data_downstream(Downstream *downstream, size_t n,
-                                   size_t padlen);
+  DefaultMemchunks *get_response_buf();
 
   // Changes stream priority of |downstream|, which is assumed to be a
   // pushed stream.
   int adjust_pushed_stream_priority(Downstream *downstream);
 
 private:
-  WriteBuffer wb_;
+  DefaultMemchunks wb_;
   std::unique_ptr<HttpsUpstream> pre_upstream_;
   DownstreamQueue downstream_queue_;
   ev_timer settings_timer_;
   ev_timer shutdown_timer_;
   ev_prepare prep_;
-  // A response buffer used to belong to Downstream object.  This is
-  // moved here when response is partially written to wb_ in
-  // send_data_callback, but before writing them all, Downstream
-  // object was destroyed.  On destruction of Downstream,
-  // pending_data_downstream_ becomes nullptr.
-  DefaultMemchunks pending_response_buf_;
-  // Downstream object whose DATA frame payload is partillay written
-  // to wb_ in send_data_callback.  This field exists to keep track of
-  // its lifetime.  When it is destroyed, its response buffer is
-  // transferred to pending_response_buf_, and this field becomes
-  // nullptr.
-  Downstream *pending_data_downstream_;
   ClientHandler *handler_;
   nghttp2_session *session_;
-  const uint8_t *data_pending_;
-  // The length of lending data to be written into wb_.  If
-  // data_pending_ is not nullptr, data_pending_ points to the data to
-  // write.  Otherwise, pending_data_downstream_->get_response_buf()
-  // if pending_data_downstream_ is not nullptr, or
-  // pending_response_buf_ holds data to write.
-  size_t data_pendinglen_;
-  size_t padding_pendinglen_;
   bool flow_control_;
   bool shutdown_handled_;
 };
