@@ -49,7 +49,8 @@ using namespace nghttp2;
 namespace shrpx {
 
 HttpsUpstream::HttpsUpstream(ClientHandler *handler)
-    : handler_(handler), current_header_length_(0),
+    : handler_(handler),
+      current_header_length_(0),
       ioctrl_(handler->get_rlimit()) {
   http_parser_init(&htp_, HTTP_REQUEST);
   htp_.data = this;
@@ -85,10 +86,13 @@ int htp_uricb(http_parser *htp, const char *data, size_t len) {
   auto downstream = upstream->get_downstream();
   auto &req = downstream->request();
 
+  auto &balloc = downstream->get_block_allocator();
+
   // We happen to have the same value for method token.
   req.method = htp->method;
 
-  if (req.fs.buffer_size() + len > get_config()->http.header_field_buffer) {
+  if (req.fs.buffer_size() + len >
+      get_config()->http.request_header_field_buffer) {
     if (LOG_ENABLED(INFO)) {
       ULOG(INFO, upstream) << "Too large URI size="
                            << req.fs.buffer_size() + len;
@@ -101,9 +105,10 @@ int htp_uricb(http_parser *htp, const char *data, size_t len) {
   req.fs.add_extra_buffer_size(len);
 
   if (req.method == HTTP_CONNECT) {
-    req.authority.append(data, len);
+    req.authority =
+        concat_string_ref(balloc, req.authority, StringRef{data, len});
   } else {
-    req.path.append(data, len);
+    req.path = concat_string_ref(balloc, req.path, StringRef{data, len});
   }
 
   return 0;
@@ -116,8 +121,9 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
   auto downstream = upstream->get_downstream();
   auto &req = downstream->request();
   auto &httpconf = get_config()->http;
+  auto &balloc = downstream->get_block_allocator();
 
-  if (req.fs.buffer_size() + len > httpconf.header_field_buffer) {
+  if (req.fs.buffer_size() + len > httpconf.request_header_field_buffer) {
     if (LOG_ENABLED(INFO)) {
       ULOG(INFO, upstream) << "Too large header block size="
                            << req.fs.buffer_size() + len;
@@ -131,7 +137,7 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
     if (req.fs.header_key_prev()) {
       req.fs.append_last_header_key(data, len);
     } else {
-      if (req.fs.num_fields() >= httpconf.max_header_fields) {
+      if (req.fs.num_fields() >= httpconf.max_request_header_fields) {
         if (LOG_ENABLED(INFO)) {
           ULOG(INFO, upstream)
               << "Too many header field num=" << req.fs.num_fields() + 1;
@@ -140,21 +146,25 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
             Downstream::HTTP1_REQUEST_HEADER_TOO_LARGE);
         return -1;
       }
-      req.fs.add_header(std::string(data, len), "");
+      auto name = http2::copy_lower(balloc, StringRef{data, len});
+      auto token = http2::lookup_token(name);
+      req.fs.add_header_token(name, StringRef{}, false, token);
     }
   } else {
     // trailer part
     if (req.fs.trailer_key_prev()) {
       req.fs.append_last_trailer_key(data, len);
     } else {
-      if (req.fs.num_fields() >= httpconf.max_header_fields) {
+      if (req.fs.num_fields() >= httpconf.max_request_header_fields) {
         if (LOG_ENABLED(INFO)) {
           ULOG(INFO, upstream)
               << "Too many header field num=" << req.fs.num_fields() + 1;
         }
         return -1;
       }
-      req.fs.add_trailer(std::string(data, len), "");
+      auto name = http2::copy_lower(balloc, StringRef{data, len});
+      auto token = http2::lookup_token(name);
+      req.fs.add_trailer_token(name, StringRef{}, false, token);
     }
   }
   return 0;
@@ -167,7 +177,8 @@ int htp_hdr_valcb(http_parser *htp, const char *data, size_t len) {
   auto downstream = upstream->get_downstream();
   auto &req = downstream->request();
 
-  if (req.fs.buffer_size() + len > get_config()->http.header_field_buffer) {
+  if (req.fs.buffer_size() + len >
+      get_config()->http.request_header_field_buffer) {
     if (LOG_ENABLED(INFO)) {
       ULOG(INFO, upstream) << "Too large header block size="
                            << req.fs.buffer_size() + len;
@@ -187,30 +198,51 @@ int htp_hdr_valcb(http_parser *htp, const char *data, size_t len) {
 } // namespace
 
 namespace {
-void rewrite_request_host_path_from_uri(Request &req, const char *uri,
+void rewrite_request_host_path_from_uri(BlockAllocator &balloc, Request &req,
+                                        const StringRef &uri,
                                         http_parser_url &u) {
   assert(u.field_set & (1 << UF_HOST));
 
-  auto &authority = req.authority;
-  authority.clear();
   // As per https://tools.ietf.org/html/rfc7230#section-5.4, we
   // rewrite host header field with authority component.
-  http2::copy_url_component(authority, &u, UF_HOST, uri);
+  auto authority = util::get_uri_field(uri.c_str(), u, UF_HOST);
   // TODO properly check IPv6 numeric address
-  if (authority.find(':') != std::string::npos) {
-    authority = '[' + authority;
-    authority += ']';
+  auto ipv6 = std::find(std::begin(authority), std::end(authority), ':') !=
+              std::end(authority);
+  auto authoritylen = authority.size();
+  if (ipv6) {
+    authoritylen += 2;
   }
   if (u.field_set & (1 << UF_PORT)) {
-    authority += ':';
-    authority += util::utos(u.port);
+    authoritylen += 1 + str_size("65535");
+  }
+  if (authoritylen > authority.size()) {
+    auto iovec = make_byte_ref(balloc, authoritylen + 1);
+    auto p = iovec.base;
+    if (ipv6) {
+      *p++ = '[';
+    }
+    p = std::copy(std::begin(authority), std::end(authority), p);
+    if (ipv6) {
+      *p++ = ']';
+    }
+
+    if (u.field_set & (1 << UF_PORT)) {
+      *p++ = ':';
+      p = util::utos(p, u.port);
+    }
+    *p = '\0';
+
+    req.authority = StringRef{iovec.base, p};
+  } else {
+    req.authority = authority;
   }
 
-  http2::copy_url_component(req.scheme, &u, UF_SCHEMA, uri);
+  req.scheme = util::get_uri_field(uri.c_str(), u, UF_SCHEMA);
 
-  std::string path;
+  StringRef path;
   if (u.field_set & (1 << UF_PATH)) {
-    http2::copy_url_component(path, &u, UF_PATH, uri);
+    path = util::get_uri_field(uri.c_str(), u, UF_PATH);
   } else if (req.method == HTTP_OPTIONS) {
     // Server-wide OPTIONS takes following form in proxy request:
     //
@@ -218,21 +250,29 @@ void rewrite_request_host_path_from_uri(Request &req, const char *uri,
     //
     // Notice that no slash after authority. See
     // http://tools.ietf.org/html/rfc7230#section-5.3.4
-    req.path = "";
+    req.path = StringRef::from_lit("");
     // we ignore query component here
     return;
   } else {
-    path = "/";
+    path = StringRef::from_lit("/");
   }
+
   if (u.field_set & (1 << UF_QUERY)) {
     auto &fdata = u.field_data[UF_QUERY];
-    path += '?';
-    path.append(uri + fdata.off, fdata.len);
+
+    if (u.field_set & (1 << UF_PATH)) {
+      auto q = util::get_uri_field(uri.c_str(), u, UF_QUERY);
+      path = StringRef{std::begin(path), std::end(q)};
+    } else {
+      path = concat_string_ref(balloc, path, StringRef::from_lit("?"),
+                               StringRef{&uri[fdata.off], fdata.len});
+    }
   }
-  if (get_config()->http2_proxy || get_config()->client_proxy) {
-    req.path = std::move(path);
+
+  if (get_config()->http2_proxy) {
+    req.path = path;
   } else {
-    req.path = http2::rewrite_clean_path(std::begin(path), std::end(path));
+    req.path = http2::rewrite_clean_path(balloc, path);
   }
 }
 } // namespace
@@ -244,6 +284,11 @@ int htp_hdrs_completecb(http_parser *htp) {
   if (LOG_ENABLED(INFO)) {
     ULOG(INFO, upstream) << "HTTP request headers completed";
   }
+
+  auto handler = upstream->get_client_handler();
+
+  handler->signal_reset_upstream_conn_rtimer();
+
   auto downstream = upstream->get_downstream();
   auto &req = downstream->request();
 
@@ -267,7 +312,7 @@ int htp_hdrs_completecb(http_parser *htp) {
     ULOG(INFO, upstream) << "HTTP request headers\n" << ss.str();
   }
 
-  if (req.fs.index_headers() != 0) {
+  if (req.fs.parse_content_length() != 0) {
     return -1;
   }
 
@@ -291,49 +336,47 @@ int htp_hdrs_completecb(http_parser *htp) {
 
   downstream->inspect_http1_request();
 
+  auto &balloc = downstream->get_block_allocator();
+
   if (method != HTTP_CONNECT) {
     http_parser_url u{};
-    // make a copy of request path, since we may set request path
-    // while we are refering to original request path.
-    auto path = req.path;
-    rv = http_parser_parse_url(path.c_str(), path.size(), 0, &u);
+    rv = http_parser_parse_url(req.path.c_str(), req.path.size(), 0, &u);
     if (rv != 0) {
       // Expect to respond with 400 bad request
       return -1;
     }
     // checking UF_HOST could be redundant, but just in case ...
     if (!(u.field_set & (1 << UF_SCHEMA)) || !(u.field_set & (1 << UF_HOST))) {
-      if (get_config()->http2_proxy || get_config()->client_proxy) {
+      if (get_config()->http2_proxy) {
         // Request URI should be absolute-form for client proxy mode
         return -1;
       }
 
       req.no_authority = true;
 
-      if (method == HTTP_OPTIONS && path == "*") {
-        req.path = "";
+      if (method == HTTP_OPTIONS && req.path == StringRef::from_lit("*")) {
+        req.path = StringRef{};
       } else {
-        req.path = http2::rewrite_clean_path(std::begin(path), std::end(path));
+        req.path = http2::rewrite_clean_path(balloc, req.path);
       }
 
       if (host) {
         req.authority = host->value;
       }
 
-      if (upstream->get_client_handler()->get_ssl()) {
-        req.scheme = "https";
+      if (handler->get_ssl()) {
+        req.scheme = StringRef::from_lit("https");
       } else {
-        req.scheme = "http";
+        req.scheme = StringRef::from_lit("http");
       }
     } else {
-      rewrite_request_host_path_from_uri(req, path.c_str(), u);
+      rewrite_request_host_path_from_uri(balloc, req, req.path, u);
     }
   }
 
   downstream->set_request_state(Downstream::HEADER_COMPLETE);
 
 #ifdef HAVE_MRUBY
-  auto handler = upstream->get_client_handler();
   auto worker = handler->get_worker();
   auto mruby_ctx = worker->get_mruby_context();
 
@@ -352,7 +395,7 @@ int htp_hdrs_completecb(http_parser *htp) {
   }
 
   rv = downstream->attach_downstream_connection(
-      upstream->get_client_handler()->get_downstream_connection(downstream));
+      handler->get_downstream_connection(downstream));
 
   if (rv != 0) {
     downstream->set_request_state(Downstream::CONNECT_FAIL);
@@ -374,6 +417,10 @@ namespace {
 int htp_bodycb(http_parser *htp, const char *data, size_t len) {
   int rv;
   auto upstream = static_cast<HttpsUpstream *>(htp->data);
+  auto handler = upstream->get_client_handler();
+
+  handler->signal_reset_upstream_conn_rtimer();
+
   auto downstream = upstream->get_downstream();
   rv = downstream->push_upload_data_chunk(
       reinterpret_cast<const uint8_t *>(data), len);
@@ -757,6 +804,7 @@ int HttpsUpstream::send_reply(Downstream *downstream, const uint8_t *body,
                               size_t bodylen) {
   const auto &req = downstream->request();
   auto &resp = downstream->response();
+  auto &balloc = downstream->get_block_allocator();
 
   auto connection_close = false;
   if (req.http_major <= 0 || (req.http_major == 1 && req.http_minor == 0)) {
@@ -776,7 +824,7 @@ int HttpsUpstream::send_reply(Downstream *downstream, const uint8_t *body,
   auto output = downstream->get_response_buf();
 
   output->append("HTTP/1.1 ");
-  output->append(http2::get_status_string(resp.http_status));
+  output->append(http2::get_status_string(balloc, resp.http_status));
   output->append("\r\n");
 
   for (auto &kv : resp.fs.headers()) {
@@ -795,6 +843,15 @@ int HttpsUpstream::send_reply(Downstream *downstream, const uint8_t *body,
     output->append("\r\n");
   }
 
+  auto &httpconf = get_config()->http;
+
+  for (auto &p : httpconf.add_response_headers) {
+    output->append(p.name);
+    output->append(": ");
+    output->append(p.value);
+    output->append("\r\n");
+  }
+
   output->append("\r\n");
 
   output->append(body, bodylen);
@@ -806,7 +863,6 @@ int HttpsUpstream::send_reply(Downstream *downstream, const uint8_t *body,
 }
 
 void HttpsUpstream::error_reply(unsigned int status_code) {
-  auto html = http::create_error_html(status_code);
   auto downstream = get_downstream();
 
   if (!downstream) {
@@ -815,6 +871,9 @@ void HttpsUpstream::error_reply(unsigned int status_code) {
   }
 
   auto &resp = downstream->response();
+  auto &balloc = downstream->get_block_allocator();
+
+  auto html = http::create_error_html(balloc, status_code);
 
   resp.http_status = status_code;
   // we are going to close connection for both frontend and backend in
@@ -825,7 +884,7 @@ void HttpsUpstream::error_reply(unsigned int status_code) {
   auto output = downstream->get_response_buf();
 
   output->append("HTTP/1.1 ");
-  auto status_str = http2::get_status_string(status_code);
+  auto status_str = http2::get_status_string(balloc, status_code);
   output->append(status_str);
   output->append("\r\nServer: ");
   output->append(get_config()->http.server_name);
@@ -886,6 +945,7 @@ int HttpsUpstream::on_downstream_header_complete(Downstream *downstream) {
 
   const auto &req = downstream->request();
   auto &resp = downstream->response();
+  auto &balloc = downstream->get_block_allocator();
 
 #ifdef HAVE_MRUBY
   if (!downstream->get_non_final_response()) {
@@ -912,13 +972,12 @@ int HttpsUpstream::on_downstream_header_complete(Downstream *downstream) {
   buf->append(".");
   buf->append(util::utos(req.http_minor));
   buf->append(" ");
-  buf->append(http2::get_status_string(resp.http_status));
+  buf->append(http2::get_status_string(balloc, resp.http_status));
   buf->append("\r\n");
 
   auto &httpconf = get_config()->http;
 
-  if (!get_config()->http2_proxy && !get_config()->client_proxy &&
-      !httpconf.no_location_rewrite) {
+  if (!get_config()->http2_proxy && !httpconf.no_location_rewrite) {
     downstream->rewrite_location_response_header(
         get_client_handler()->get_upstream_scheme());
   }
@@ -987,7 +1046,7 @@ int HttpsUpstream::on_downstream_header_complete(Downstream *downstream) {
     }
   }
 
-  if (!get_config()->http2_proxy && !get_config()->client_proxy) {
+  if (!get_config()->http2_proxy) {
     buf->append("Server: ");
     buf->append(httpconf.server_name);
     buf->append("\r\n");
@@ -1013,15 +1072,17 @@ int HttpsUpstream::on_downstream_header_complete(Downstream *downstream) {
       buf->append((*via).value);
       buf->append(", ");
     }
-    buf->append(
-        http::create_via_header_value(resp.http_major, resp.http_minor));
+    std::array<char, 16> viabuf;
+    auto end = http::create_via_header_value(viabuf.data(), resp.http_major,
+                                             resp.http_minor);
+    buf->append(viabuf.data(), end - std::begin(viabuf));
     buf->append("\r\n");
   }
 
   for (auto &p : httpconf.add_response_headers) {
-    buf->append(p.first);
+    buf->append(p.name);
     buf->append(": ");
-    buf->append(p.second);
+    buf->append(p.value);
     buf->append("\r\n");
   }
 
@@ -1142,8 +1203,7 @@ fail:
   return 0;
 }
 
-int HttpsUpstream::initiate_push(Downstream *downstream, const char *uri,
-                                 size_t len) {
+int HttpsUpstream::initiate_push(Downstream *downstream, const StringRef &uri) {
   return 0;
 }
 
