@@ -64,7 +64,7 @@ void drop_privileges(
 #endif // HAVE_NEVERBLEED
     ) {
   if (getuid() == 0 && get_config()->uid != 0) {
-    if (initgroups(get_config()->user.get(), get_config()->gid) != 0) {
+    if (initgroups(get_config()->user.c_str(), get_config()->gid) != 0) {
       auto error = errno;
       LOG(FATAL) << "Could not change supplementary groups: "
                  << strerror(error);
@@ -86,7 +86,7 @@ void drop_privileges(
     }
 #ifdef HAVE_NEVERBLEED
     if (nb) {
-      neverbleed_setuidgid(nb, get_config()->user.get(), 1);
+      neverbleed_setuidgid(nb, get_config()->user.c_str(), 1);
     }
 #endif // HAVE_NEVERBLEED
   }
@@ -389,19 +389,12 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
 
   ConnectionHandler conn_handler(loop);
 
-  if (wpconf->server_fd6 != -1) {
-    conn_handler.set_acceptor6(
-        make_unique<AcceptHandler>(wpconf->server_fd6, &conn_handler));
+  for (auto &addr : get_config()->conn.listener.addrs) {
+    conn_handler.add_acceptor(make_unique<AcceptHandler>(&addr, &conn_handler));
   }
-  if (wpconf->server_fd != -1) {
-    conn_handler.set_acceptor(
-        make_unique<AcceptHandler>(wpconf->server_fd, &conn_handler));
-  }
-
-  auto &upstreamconf = get_config()->conn.upstream;
 
 #ifdef HAVE_NEVERBLEED
-  if (!upstreamconf.no_tls || ssl::downstream_tls_enabled()) {
+  if (ssl::upstream_tls_enabled() || ssl::downstream_tls_enabled()) {
     std::array<char, NEVERBLEED_ERRBUF_SIZE> errbuf;
     auto nb = make_unique<neverbleed_t>();
     if (neverbleed_init(nb.get(), errbuf.data()) != 0) {
@@ -425,13 +418,24 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
 
 #endif // HAVE_NEVERBLEED
 
+  MemchunkPool mcpool;
+
   ev_timer renew_ticket_key_timer;
-  if (!upstreamconf.no_tls) {
+  if (ssl::upstream_tls_enabled()) {
     auto &ticketconf = get_config()->tls.ticket;
+    auto &memcachedconf = ticketconf.memcached;
+
+    if (!memcachedconf.host.empty()) {
+      SSL_CTX *ssl_ctx = nullptr;
+
+      if (memcachedconf.tls) {
+        ssl_ctx = conn_handler.create_tls_ticket_key_memcached_ssl_ctx();
+      }
 
-    if (ticketconf.memcached.host) {
       conn_handler.set_tls_ticket_key_memcached_dispatcher(
-          make_unique<MemcachedDispatcher>(&ticketconf.memcached.addr, loop));
+          make_unique<MemcachedDispatcher>(
+              &ticketconf.memcached.addr, loop, ssl_ctx,
+              StringRef{memcachedconf.host}, &mcpool));
 
       ev_timer_init(&renew_ticket_key_timer, memcached_get_ticket_key_cb, 0.,
                     0.);
@@ -516,7 +520,7 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
   ipcev.data = &conn_handler;
   ev_io_start(loop, &ipcev);
 
-  if (!upstreamconf.no_tls && !get_config()->tls.ocsp.disabled) {
+  if (ssl::upstream_tls_enabled() && !get_config()->tls.ocsp.disabled) {
     conn_handler.proceed_next_cert_ocsp();
   }
 
