@@ -59,6 +59,13 @@ struct TLSSessionInfo {
 
 TLSSessionInfo *get_tls_session_info(TLSSessionInfo *tls_info, SSL *ssl);
 
+// Returns true iff the negotiated protocol is TLSv1.2.
+bool check_http2_tls_version(SSL *ssl);
+
+// Returns true iff the negotiated cipher suite is in HTTP/2 cipher
+// black list.
+bool check_http2_cipher_black_list(SSL *ssl);
+
 // Returns true if SSL/TLS requirement for HTTP/2 is fulfilled.
 // To fulfill the requirement, the following 2 terms must be hold:
 //
