@@ -25,6 +25,10 @@
 #ifndef FAILMALLOC_TEST_H
 #define FAILMALLOC_TEST_H
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif /* HAVE_CONFIG_H */
+
 void test_nghttp2_session_send(void);
 void test_nghttp2_session_recv(void);
 void test_nghttp2_frame(void);
